@@ -1,32 +1,34 @@
 /*
- * mon_procd
- * Author: Hongjie Yang <hongjie@us.ibm.com>
+ * mon_procd - Write process data to the z/VM monitor stream
  *
- * Copyright IBM Corp. 2007.
+ * Copyright IBM Corp. 2007, 2017
  *
- * Daemon that writes process data to the z/VM monitor stream.
+ * s390-tools is free software; you can redistribute it and/or modify
+ * it under the terms of the MIT license. See LICENSE for details.
  */
-#include <unistd.h>
-#include <sys/vfs.h>
-#include <stdlib.h>
-#include <stdio.h>
-#include <string.h>
-#include <fcntl.h>
-#include <syslog.h>
-#include <errno.h>
-#include <time.h>
-#include <stdarg.h>
-#include <linux/types.h>
-#include <sys/statvfs.h>
-#include <sys/stat.h>
-#include <sys/dir.h>
-#include <getopt.h>
-#include <signal.h>
-#include <utmp.h>
-#include <pwd.h>
-#include <grp.h>
+
 #include <ctype.h>
 #include <dirent.h>
+#include <errno.h>
+#include <fcntl.h>
+#include <getopt.h>
+#include <grp.h>
+#include <linux/types.h>
+#include <pwd.h>
+#include <signal.h>
+#include <stdarg.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+#include <sys/dir.h>
+#include <sys/stat.h>
+#include <sys/statvfs.h>
+#include <sys/vfs.h>
+#include <syslog.h>
+#include <time.h>
+#include <unistd.h>
+#include <utmp.h>
+
 #include "mon_procd.h"
 
 struct name_lens_t {
@@ -107,17 +109,18 @@ static void procd_open_monwriter(void)
 /*
  * Store daemon's pid so it can be stopped
  */
-static void store_pid(void)
+static int store_pid(void)
 {
 	FILE *f = fopen(pid_file, "w");
 
 	if (!f) {
 		syslog(LOG_ERR, "cannot open pid file %s: %s", pid_file,
 		       strerror(errno));
-		exit(1);
+		return -1;
 	}
 	fprintf(f, "%d\n", getpid());
 	fclose(f);
+	return 0;
 }
 
 /*
@@ -129,6 +132,9 @@ static void stop_unused(int curr_max, int prev_max)
 	int i;
 
 	mw_hdrp = (struct monwrite_hdr *)mon_record;
+	mw_hdrp->applid = PROCD_APPLID;
+	mw_hdrp->record_num = TASK_FLAG;
+	mw_hdrp->hdrlen = sizeof(struct monwrite_hdr);
 	mw_hdrp->mon_function = MONWRITE_STOP_INTERVAL;
 	mw_hdrp->datalen = 0;
 	for (i = 0; i < prev_max - curr_max; i += 2) {
@@ -194,7 +200,7 @@ static void procd_write_ent(void *entry, int size, char flag)
 }
 
 /*
- * Open and read a file into a buffer
+ * Open and read a file into a buffer, terminated with buf[size/num] = '\0'
 */
 static int read_file(char *fname, char *buf, int size)
 {
@@ -209,6 +215,7 @@ static int read_file(char *fname, char *buf, int size)
 		close(fp);
 		return -1;
 	}
+	buf[num] = '\0';
 
 	close(fp);
 	return num;
@@ -325,9 +332,13 @@ static void read_cpu(void)
 	if (read_file("/proc/stat", buf, sizeof(buf) - 1) <= 0)
 		return;
 
+	u = n = s = i = w = x = y = z = 0;
 	temp = strstr(buf, "cpu");
-	sscanf(temp, "cpu %Lu %Lu %Lu %Lu %Lu %Lu %Lu %Lu",
-		&u, &n, &s, &i, &w, &x, &y, &z);
+	if (temp)
+		sscanf(temp, "cpu %Lu %Lu %Lu %Lu %Lu %Lu %Lu %Lu",
+			&u, &n, &s, &i, &w, &x, &y, &z);
+	else
+		syslog(LOG_ERR, "no cpu in /proc/stat\n");
 	cpudata.usr = (__u64)u;
 	cpudata.nice = (__u64)n;
 	cpudata.sys = (__u64)s;
@@ -351,18 +362,37 @@ static void read_mem(void)
 	if (read_file("/proc/meminfo", buf, sizeof(buf) - 1) <= 0)
 		return;
 
+	mtotal = mfree = mbuf = stotal = sfree = scached = 0;
 	temp = strstr(buf, "MemTotal");
-	sscanf(temp, "MemTotal: %Lu kB", &mtotal);
+	if (temp)
+		sscanf(temp, "MemTotal: %Lu kB", &mtotal);
+	else
+		syslog(LOG_ERR, "no MemTotal in /proc/meminfo\n");
 	temp = strstr(buf, "MemFree");
-	sscanf(temp, "MemFree: %Lu kB", &mfree);
+	if (temp)
+		sscanf(temp, "MemFree: %Lu kB", &mfree);
+	else
+		syslog(LOG_ERR, "no MemFree in /proc/meminfo\n");
 	temp = strstr(buf, "Buffers");
-	sscanf(temp, "Buffers: %Lu kB", &mbuf);
+	if (temp)
+		sscanf(temp, "Buffers: %Lu kB", &mbuf);
+	else
+		syslog(LOG_ERR, "no Buffers in /proc/meminfo\n");
 	temp = strstr(buf, "SwapTotal");
-	sscanf(temp, "SwapTotal: %Lu kB", &stotal);
+	if (temp)
+		sscanf(temp, "SwapTotal: %Lu kB", &stotal);
+	else
+		syslog(LOG_ERR, "no SwapTotal in /proc/meminfo\n");
 	temp = strstr(buf, "SwapFree");
-	sscanf(temp, "SwapFree: %Lu kB", &sfree);
+	if (temp)
+		sscanf(temp, "SwapFree: %Lu kB", &sfree);
+	else
+		syslog(LOG_ERR, "no SwapFree in /proc/meminfo\n");
 	temp = strstr(buf, "Cached");
-	sscanf(temp, "Cached: %Lu kB", &scached);
+	if (temp)
+		sscanf(temp, "Cached: %Lu kB", &scached);
+	else
+		syslog(LOG_ERR, "no Cached in /proc/meminfo\n");
 
 	proc_sum.mem.total = (__u64)mtotal;
 	proc_sum.mem.free = (__u64)mfree;
@@ -384,14 +414,27 @@ static void read_vmem(void)
 	if (read_file("/proc/vmstat", buf, sizeof(buf) - 1) <= 0)
 		return;
 
+	pgin = pgout = swpin = swpout = 0;
 	temp = strstr(buf, "pgpgin");
-	sscanf(temp, "pgpgin %Lu", &pgin);
+	if (temp)
+		sscanf(temp, "pgpgin %Lu", &pgin);
+	else
+		syslog(LOG_ERR, "no pgpgin in /proc/vmstat\n");
 	temp = strstr(buf, "pgpgout");
-	sscanf(temp, "pgpgout %Lu", &pgout);
+	if (temp)
+		sscanf(temp, "pgpgout %Lu", &pgout);
+	else
+		syslog(LOG_ERR, "no pgpgout in /proc/vmstat\n");
 	temp = strstr(buf, "pswpin");
-	sscanf(temp, "pswpin %Lu", &swpin);
+	if (temp)
+		sscanf(temp, "pswpin %Lu", &swpin);
+	else
+		syslog(LOG_ERR, "no pswpin in /proc/vmstat\n");
 	temp = strstr(buf, "pswpout");
-	sscanf(temp, "pswpout %Lu", &swpout);
+	if (temp)
+		sscanf(temp, "pswpout %Lu", &swpout);
+	else
+		syslog(LOG_ERR, "no pswpout in /proc/vmstat\n");
 
 	proc_sum.mem.pgpgin = (__u64)(pgin << pg_to_kb_shift);
 	proc_sum.mem.pgpgout = (__u64)(pgout << pg_to_kb_shift);
@@ -451,10 +494,17 @@ static int read_status(struct task_t *task)
 	if (read_file(fname, buf, sizeof(buf) - 1) == -1)
 		return 0;
 
+	ruid = euid = egid = 0;
 	temp = strstr(buf, "Uid");
-	sscanf(temp, "Uid: %d %d", &ruid, &euid);
+	if (temp)
+		sscanf(temp, "Uid: %d %d", &ruid, &euid);
+	else
+		syslog(LOG_ERR, "no Uid in /proc/%u/status\n", task->pid);
 	temp = strstr(buf, "Gid");
-	sscanf(temp, "Gid: %*d %d", &egid);
+	if (temp)
+		sscanf(temp, "Gid: %*d %d", &egid);
+	else
+		syslog(LOG_ERR, "no Gid in /proc/%u/status\n", task->pid);
 	task->euid = (__u16)euid;
 
 	lenp = mon_record + sizeof(struct monwrite_hdr);
@@ -848,41 +898,64 @@ static void read_tasks(void)
  */
 static void procd_daemonize(void)
 {
+	int pipe_fds[2], startup_rc = 1;
 	pid_t pid;
 
+	if (pipe(pipe_fds) == -1) {
+		syslog(LOG_ERR, "pipe error: %s\n", strerror(errno));
+		exit(1);
+	}
+
 	/* Fork off the parent process */
 	pid = fork();
 	if (pid < 0) {
 		syslog(LOG_ERR, "fork error: %s\n", strerror(errno));
 		exit(1);
 	}
-	if (pid > 0)
-		exit(0);
+	if (pid > 0) {
+		/* Wait for startup return code from daemon */
+		if (read(pipe_fds[0], &startup_rc, sizeof(startup_rc)) == -1)
+			syslog(LOG_ERR, "pipe read error: %s\n", strerror(errno));
+		/* With startup_rc == 0, pid file was written at this point */
+		exit(startup_rc);
+	}
 
 	/* Change the file mode mask */
 	umask(0);
 
-	/* Store daemon pid */
-	store_pid();
 	/* Catch SIGINT and SIGTERM to clean up pid file on exit */
 	procd_handle_signals();
 
 	/* Create a new SID for the child process */
 	if (setsid() < 0) {
 		syslog(LOG_ERR, "setsid error: %s\n",  strerror(errno));
-		exit(1);
+		goto notify_parent;
 	}
 
 	/* Change the current working directory */
 	if (chdir("/") < 0) {
 		syslog(LOG_ERR, "chdir error: %s\n",  strerror(errno));
-		exit(1);
+		goto notify_parent;
 	}
 
 	/* Close out the standard file descriptors */
 	close(STDIN_FILENO);
 	close(STDOUT_FILENO);
 	close(STDERR_FILENO);
+
+	/* Store daemon pid */
+	if (store_pid() < 0)
+		goto notify_parent;
+	startup_rc = 0;
+
+notify_parent:
+	/* Inform waiting parent about startup return code */
+	if (write(pipe_fds[1], &startup_rc, sizeof(startup_rc)) == -1) {
+		syslog(LOG_ERR, "pipe write error: %s\n", strerror(errno));
+		exit(1);
+	}
+	if (startup_rc != 0)
+		exit(startup_rc);
 }
 
 static int procd_do_work(void)
@@ -953,7 +1026,7 @@ static int parse_options(int argc, char **argv)
 			exit(0);
 		case 'v':
 			printf( "mon_procd: version %s\n", RELEASE_STRING);
-			printf("Copyright IBM Corp. 2007\n");
+			printf("Copyright IBM Corp. 2007, 2017\n");
 			exit(0);
 		case 'a':
 			attach = 1;
