@@ -1,8 +1,5 @@
 /*
- * vmcp
- * Author: Christian Borntraeger <cborntra@de.ibm.com>
- *
- * Copyright IBM Corp. 2005, 2006.
+ * vmcp - Send commands to the z/VM control program
  *
  * Tool for accessing the control program of z/VM using the
  * kernel module vmcp
@@ -10,10 +7,15 @@
  *	0: everything was fine                      (VMCP_OK)
  *	1: CP returned a nn zero response code      (VMCP_CP)
  *	2: the response buffer was not large enough (VMCP_BUF)
- *	3: an internal Linux error occured          (VMCP_LIN)
+ *	3: an internal Linux error occurred         (VMCP_LIN)
  *	4: invalid options                          (VMCP_OPT)
  *
  * CREDITS: The idea is based on cpint of Neale Fergusson
+ *
+ * Copyright IBM Corp. 2005, 2017
+ *
+ * s390-tools is free software; you can redistribute it and/or modify
+ * it under the terms of the MIT license. See LICENSE for details.
  */
 
 #include <ctype.h>
@@ -23,10 +25,11 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <unistd.h>
 #include <sys/ioctl.h>
-#include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/types.h>
+#include <unistd.h>
+
 #include "vmcp.h"
 
 static int keep_case = 0;
@@ -40,8 +43,8 @@ static void print_help(const char *name)
 
 static void print_version(const char *name)
 {
-	printf("%s: z/VM CP command interface version %s.\n"
-	       "Copyright IBM Corp. 2005, 2006.\n",
+	printf("%s: z/VM CP command interface version %s\n"
+	       "Copyright IBM Corp. 2005, 2017\n",
 	       name, RELEASE_STRING);
 }
 
@@ -152,8 +155,10 @@ static ssize_t read_buffer(int fd, char *buf, size_t count)
 	ssize_t ret;
 	ssize_t done;
 
-	for (done = 0; done < count; done += ret) {
+	for (done = 0; done < (ssize_t) count; done += ret) {
 		ret = read(fd, &buf[done], count - done);
+		if (ret == -1 && errno == EINTR)
+			continue;
 		if (ret == -1)
 			return -1;
 		if (ret == 0)
@@ -169,8 +174,10 @@ static ssize_t write_buffer(int fd, const char *buf, size_t count)
 	ssize_t ret;
 	ssize_t done;
 
-	for (done = 0; done < count; done += ret) {
+	for (done = 0; done < (ssize_t) count; done += ret) {
 		ret = write(fd, &buf[done], count - done);
+		if (ret == -1 && errno == EINTR)
+			continue;
 		if (ret == -1)
 			return -1;
 		if (ret == 0)
