@@ -2,22 +2,23 @@
  * s390-tools/zipl/include/bootmap.h
  *   Functions to build the bootmap file.
  *
- * Copyright IBM Corp. 2001, 2006.
+ * Copyright IBM Corp. 2001, 2017
+ *
+ * s390-tools is free software; you can redistribute it and/or modify
+ * it under the terms of the MIT license. See LICENSE for details.
  *
- * Author(s): Carsten Otte <cotte@de.ibm.com>
- *            Peter Oberparleiter <Peter.Oberparleiter@de.ibm.com>
  */
 
 #ifndef BOOTMAP_H
 #define BOOTMAP_H
 
-#include "zipl.h"
-
-#include "job.h"
 #include "disk.h"
+#include "job.h"
+#include "zipl.h"
 
 
 int bootmap_create(struct job_data* job, disk_blockptr_t* program_table,
+		   disk_blockptr_t *scsi_dump_sb_blockptr,
 		   disk_blockptr_t** stage2_list, blocknum_t* stage2_count,
 		   char** device, struct disk_info** new_info);
 void bootmap_store_blockptr(void* buffer, disk_blockptr_t* ptr,
