@@ -1,30 +1,29 @@
 /*
- * s390-tools/zipl/src/job.c
- *   Functions and data structures representing the actual 'job' that the
- *   user wants us to execute.
+ * zipl - zSeries Initial Program Loader tool
  *
- * Copyright IBM Corp. 2001, 2009.
+ * Functions and data structures representing the actual 'job' that the
+ * user wants us to execute
  *
- * Author(s): Carsten Otte <cotte@de.ibm.com>
- *            Peter Oberparleiter <Peter.Oberparleiter@de.ibm.com>
+ * Copyright IBM Corp. 2001, 2017
+ *
+ * s390-tools is free software; you can redistribute it and/or modify
+ * it under the terms of the MIT license. See LICENSE for details.
  */
 
-#include "job.h"
-
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <unistd.h>
+#include <errno.h>
+#include <getopt.h>
 #include <stdio.h>
 #include <stdlib.h>
-#include <getopt.h>
 #include <string.h>
-#include <errno.h>
+#include <sys/stat.h>
+#include <sys/types.h>
+#include <unistd.h>
 
 #include "error.h"
+#include "job.h"
 #include "misc.h"
 #include "scan.h"
 
-
 /* Command line options */
 static struct option options[] = {
 	{ "config",		required_argument,	NULL, 'c'},
@@ -51,11 +50,12 @@ static struct option options[] = {
 	{ "tape",		required_argument,	NULL, 'T'},
 	{ "dry-run",		no_argument,		NULL, '0'},
 	{ "force",		no_argument,		NULL, 'f'},
+	{ "kdump",		required_argument,	NULL, 'k'},
 	{ NULL,			0,			NULL, 0 }
 };
 
 /* Command line option abbreviations */
-static const char option_string[] = "-c:t:i:r:p:P:d:D:M:s:m:hHnVvaT:f";
+static const char option_string[] = "-c:t:i:r:p:P:d:D:M:s:m:hHnVvaT:fk:";
 
 struct command_line {
 	char* data[SCAN_KEYWORD_NUM];
@@ -110,9 +110,9 @@ get_command_line(int argc, char* argv[], struct command_line* line)
 					  optarg);
 			break;
 		case 'D':
-			is_keyword = 1;
-			rc = store_option(&cmdline, scan_keyword_dumptofs,
-					  optarg);
+			error_reason("dumptofs has been deprecated, use "
+				     "--dump instead");
+			rc = -1;
 			break;
 		case 'M':
 			is_keyword = 1;
@@ -184,6 +184,11 @@ get_command_line(int argc, char* argv[], struct command_line* line)
 			rc = store_option(&cmdline, scan_keyword_tape,
 					  optarg);
 			break;
+		case 'k':
+			is_keyword = 1;
+			rc = store_option(&cmdline, scan_keyword_kdump,
+					  optarg);
+			break;
 		case 'c':
 			if (cmdline.config != NULL) {
 				error_reason("Option 'config' specified more "
@@ -338,12 +343,16 @@ free_segment_data(struct job_segment_data* data)
 
 
 static void
-free_dump_fs_data(struct job_dump_fs_data* data)
+free_dump_data(struct job_dump_data* data)
 {
-	if (data->partition != NULL)
-		free(data->partition);
+	if (data->device != NULL)
+		free(data->device);
+	if (data->image != NULL)
+		free(data->image);
 	if (data->parmline != NULL)
 		free(data->parmline);
+	if (data->ramdisk != NULL)
+		free(data->ramdisk);
 }
 
 
@@ -360,9 +369,9 @@ free_menu_data(struct job_menu_data* data)
 			case job_ipl:
 				free_ipl_data(&data->entry[i].data.ipl);
 				break;
-			case job_dump_fs:
-				free_dump_fs_data(
-					&data->entry[i].data.dump_fs);
+			case job_dump_partition:
+				free_dump_data(
+					&data->entry[i].data.dump);
 				break;
 			default:
 				break;
@@ -373,14 +382,6 @@ free_menu_data(struct job_menu_data* data)
 }
 
 
-void
-free_dump_data(struct job_dump_data* data)
-{
-	if (data->device != NULL)
-		free(data->device);
-}
-
-
 void
 free_mvdump_data(struct job_mvdump_data* data)
 {
@@ -418,9 +419,6 @@ job_free(struct job_data* job)
 	case job_mvdump:
 		free_mvdump_data(&job->data.mvdump);
 		break;
-	case job_dump_fs:
-		free_dump_fs_data(&job->data.dump_fs);
-		break;
 	case job_ipl_tape:
 		free_ipl_tape_data(&job->data.ipl_tape);
 		break;
@@ -431,6 +429,360 @@ job_free(struct job_data* job)
 }
 
 
+struct component_loc {
+	char *name;
+	address_t *addrp;
+	size_t size;
+	off_t align;
+};
+
+static int
+set_cl_element(struct component_loc *cl, char *name, char *filename,
+	       address_t *addrp, size_t size, off_t off, off_t align)
+{
+	struct stat stats;
+
+	cl->name = name;
+	cl->addrp = addrp;
+	cl->align = align;
+	if (size != 0) {
+		/* The loader works on blocks, so alignment is required */
+		cl->size = ALIGN(size, MAXIMUM_PHYSICAL_BLOCKSIZE);
+		return 0;
+	}
+	/* Get size */
+	if (stat(filename, &stats)) {
+		error_reason(strerror(errno));
+		error_text("Could not get information for file '%s'", filename);
+		return -1;
+	}
+	cl->size = ALIGN(stats.st_size - off, MAXIMUM_PHYSICAL_BLOCKSIZE);
+	return 0;
+}
+
+static void
+sort_cl_array(struct component_loc* cl, int elements)
+{
+	int i, j, min;
+	struct component_loc swap;
+
+	/* Selection sort keeping sequence */
+	for (i = 0; i < elements - 1; i++) {
+		min = i;
+		for (j = i + 1; j < elements; j++) {
+			if (*(cl[j].addrp) == UNSPECIFIED_ADDRESS)
+				continue;
+			if ((*(cl[min].addrp) == UNSPECIFIED_ADDRESS) ||
+			    (*(cl[j].addrp) < *(cl[min].addrp)))
+				min = j;
+		}
+		if (i != min) {
+			swap = cl[i];
+			cl[i] = cl[min];
+			for (j = min; j > i + 1; j--)
+				cl[j] = cl[j - 1];
+			cl[i + 1] = swap;
+		}
+	}
+}
+
+static int
+get_ipl_components(struct job_ipl_data *ipl, struct component_loc **clp,
+		   int *nump)
+{
+	struct component_loc *cl;
+	int num;
+	int rc;
+
+	/* Get memory for image, parmline, ramdisk */
+	cl = misc_calloc(3, sizeof(struct component_loc));
+	if (cl == NULL)
+		return -1;
+	/* Fill in component data */
+	num = 0;
+	rc = set_cl_element(&cl[num++], "kernel image", ipl->image,
+			    &ipl->image_addr, 0, 0x10000,
+			    MAXIMUM_PHYSICAL_BLOCKSIZE);
+	if (rc)
+		goto error;
+	if (ipl->parmline) {
+		rc = set_cl_element(&cl[num++], "parmline", NULL,
+				    &ipl->parm_addr, MAXIMUM_PARMLINE_SIZE, 0,
+				    MAXIMUM_PHYSICAL_BLOCKSIZE);
+		if (rc)
+			goto error;
+	}
+	if (ipl->ramdisk) {
+		rc = set_cl_element(&cl[num++], "initial ramdisk", ipl->ramdisk,
+			&ipl->ramdisk_addr, 0, 0, 0x10000);
+		if (rc)
+			goto error;
+	}
+
+	*clp = cl;
+	*nump = num;
+	return 0;
+error:
+	free(cl);
+	return rc;
+}
+
+
+static int
+get_dump_components(struct job_dump_data *dump,
+		       struct component_loc **clp, int *nump)
+{
+	struct component_loc *cl;
+	int num;
+	int rc;
+
+	/* Get memory for image, parmline, ramdisk, loader */
+	cl = misc_calloc(3, sizeof(struct component_loc));
+	if (cl == NULL)
+		return -1;
+	/* Fill in component data */
+	num = 0;
+	rc = set_cl_element(&cl[num++], "kernel image", dump->image,
+			    &dump->image_addr, 0, 0x10000,
+			    MAXIMUM_PHYSICAL_BLOCKSIZE);
+	if (rc)
+		goto error;
+	rc = set_cl_element(&cl[num++], "parmline", NULL, &dump->parm_addr,
+			    MAXIMUM_PARMLINE_SIZE, 0,
+			    MAXIMUM_PHYSICAL_BLOCKSIZE);
+	if (rc)
+		goto error;
+	if (dump->ramdisk) {
+		rc = set_cl_element(&cl[num++], "initial ramdisk",
+				    dump->ramdisk, &dump->ramdisk_addr,
+				    0, 0, 0x10000);
+		if (rc)
+			goto error;
+	}
+
+	*clp = cl;
+	*nump = num;
+	return 0;
+error:
+	free(cl);
+	return rc;
+}
+
+
+static int
+get_ipl_tape_components(struct job_ipl_tape_data *ipl_tape,
+			struct component_loc **clp, int *nump)
+{
+	struct component_loc *cl;
+	int num;
+	int rc;
+
+	/* Get memory for image, parmline, ramdisk */
+	cl = misc_calloc(3, sizeof(struct component_loc));
+	if (cl == NULL)
+		return -1;
+	/* Fill in component data */
+	num = 0;
+	rc = set_cl_element(&cl[num++], "kernel image", ipl_tape->image,
+			    &ipl_tape->image_addr, 0, 0x10000,
+			    MAXIMUM_PHYSICAL_BLOCKSIZE);
+	if (rc)
+		goto error;
+	/* Enhance the image size value for tape loader specific logic */
+	cl[0].size = ALIGN(cl[0].size + 0x100, MAXIMUM_PHYSICAL_BLOCKSIZE);
+	if (ipl_tape->parmline) {
+		rc = set_cl_element(&cl[num++], "parmline", NULL,
+				    &ipl_tape->parm_addr, MAXIMUM_PARMLINE_SIZE,
+				    0, MAXIMUM_PHYSICAL_BLOCKSIZE);
+		if (rc)
+			goto error;
+	}
+	if (ipl_tape->ramdisk) {
+		rc = set_cl_element(&cl[num++], "initial ramdisk",
+				    ipl_tape->ramdisk, &ipl_tape->ramdisk_addr,
+				    0, 0, 0x10000);
+		if (rc)
+			goto error;
+	}
+	*clp = cl;
+	*nump = num;
+	return 0;
+error:
+	free(cl);
+	return rc;
+}
+
+
+static int
+check_component_address_data(struct component_loc *cl, int num, char *name,
+			     unsigned long address_limit)
+{
+	int i;
+
+	/* Check for address limit */
+	for (i = 0; i < num; i++) {
+		if (*cl[i].addrp == UNSPECIFIED_ADDRESS)
+			continue;
+		if (*cl[i].addrp + cl[i].size > address_limit) {
+			if (name != NULL)
+				error_text("Section '%s'", name);
+			error_reason("Component '%s' exceeds available address "
+				     "space (limit is 0x%08x)", cl[i].name,
+				     address_limit);
+			return -1;
+		}
+		if (*cl[i].addrp < MINIMUM_ADDRESS) {
+			if (name != NULL)
+				error_text("Section '%s'", name);
+			error_reason("Component '%s' falls below available "
+				     "address space (limit is 0x%08x)",
+				     cl[i].name, MINIMUM_ADDRESS);
+			return -1;
+		}
+	}
+	/* Check for overlap */
+	for (i = 0; i < num - 1; i++) {
+		if (*cl[i].addrp == UNSPECIFIED_ADDRESS ||
+		    *cl[i + 1].addrp == UNSPECIFIED_ADDRESS)
+			continue;
+		if (*cl[i].addrp + cl[i].size > *cl[i + 1].addrp) {
+			if (name != NULL)
+				error_text("Section '%s'", name);
+			error_reason("Components '%s' and '%s' overlap",
+				     cl[i].name, cl[i + 1].name);
+			return -1;
+		}
+	}
+	return 0;
+}
+
+
+static int
+finalize_component_address_data(struct component_loc *cl, int num,
+				unsigned long address_limit)
+{
+	struct component_loc swap;
+	address_t addr;
+	int i;
+	int j;
+	int k;
+
+	/* Calculate unspecified addresses */
+	for (i = 0; i < num; i++) {
+		if (*cl[i].addrp != UNSPECIFIED_ADDRESS)
+			continue;
+		for (j = -1; j < i; j++) {
+			if (j < 0) {
+				/* Try address before first component */
+				addr = MINIMUM_ADDRESS;
+			} else {
+				/* Try address after component j */
+				addr = *cl[j].addrp + cl[j].size;
+				if (addr < MINIMUM_ADDRESS)
+					addr = MINIMUM_ADDRESS;
+			}
+			addr = ALIGN(addr, cl[i].align);
+			if (addr + cl[i].size > address_limit) {
+				error_text("Could not fit component '%s' into "
+					   "available address space",
+					   cl[i].name);
+				return -1;
+			}
+			/* Check for enough room */
+			if (*cl[j + 1].addrp != UNSPECIFIED_ADDRESS &&
+			    addr + cl[i].size > *cl[j + 1].addrp)
+				continue;
+			*cl[i].addrp = addr;
+			/* If there is no next component we are done */
+			if (i == j + 1)
+				break;
+			/* Restore sort order */
+			swap = cl[j + 1];
+			cl[j + 1] = cl[i];
+			for (k = i; k > j + 2; k--)
+				cl[k] = cl[k - 1];
+			cl[j + 2] = swap;
+			break;
+		}
+	}
+	return 0;
+}
+
+
+static int
+finalize_ipl_address_data(struct job_ipl_data* ipl, char *name)
+{
+	unsigned long address_limit;
+	struct component_loc *cl;
+	int num;
+	int rc;
+
+	address_limit = ipl->is_kdump ? ADDRESS_LIMIT_KDUMP : ADDRESS_LIMIT;
+	rc = get_ipl_components(ipl, &cl, &num);
+	if (rc)
+		return rc;
+	sort_cl_array(cl, num);
+	rc = check_component_address_data(cl, num, name, address_limit);
+	if (rc)
+		goto out_free;
+	rc = finalize_component_address_data(cl, num, address_limit);
+out_free:
+	free(cl);
+	return rc;
+}
+
+
+static int
+finalize_dump_address_data(struct job_dump_data* dump, char *name)
+{
+	struct component_loc *cl;
+	int num;
+	int rc;
+
+	rc = get_dump_components(dump, &cl, &num);
+	if (rc)
+		return rc;
+	sort_cl_array(cl, num);
+	rc = check_component_address_data(cl, num, name, ADDRESS_LIMIT);
+	if (rc)
+		goto out_free;
+	rc = finalize_component_address_data(cl, num, ADDRESS_LIMIT);
+out_free:
+	free(cl);
+	return rc;
+}
+
+
+static int
+finalize_ipl_tape_address_data(struct job_ipl_tape_data* ipl_tape, char *name)
+{
+	struct component_loc *cl;
+	int num;
+	int rc;
+	int i;
+
+	rc = get_ipl_tape_components(ipl_tape, &cl, &num);
+	if (rc)
+		return rc;
+	sort_cl_array(cl, num);
+	rc = check_component_address_data(cl, num, name, ADDRESS_LIMIT);
+	if (rc)
+		goto out_free;
+	rc = finalize_component_address_data(cl, num, ADDRESS_LIMIT);
+	/* The loader needs a ramdisk_addr anyway */
+	if (ipl_tape->ramdisk_addr == 0) {
+		for (i = 0; i < num; i++) {
+			if (*cl[i].addrp == ipl_tape->image_addr)
+				ipl_tape->ramdisk_addr =
+					*cl[i].addrp + cl[i].size;
+		}
+	}
+out_free:
+	free(cl);
+	return rc;
+}
+
+
 static int
 check_job_ipl_data(struct job_ipl_data *ipl, char* name)
 {
@@ -460,7 +812,7 @@ check_job_ipl_data(struct job_ipl_data *ipl, char* name)
 			return rc;
 		}
 	}
-	return 0;
+	return finalize_ipl_address_data(ipl, name);
 }
 
 
@@ -507,26 +859,34 @@ check_job_dump_data(struct job_dump_data* dump, char* name)
 }
 
 
-static int
-check_job_dump_fs_data(struct job_dump_fs_data* dump_fs, char* name)
+int
+check_job_dump_images(struct job_dump_data* dump, char* name)
 {
 	int rc;
-
-	if (dump_fs->partition != NULL) {
-		rc = misc_check_writable_device(dump_fs->partition, 1, 0);
-		if (rc) {
-			if (name == NULL) {
-				error_text("Dump partition '%s'",
-					dump_fs->partition);
-			} else {
-				error_text("Dump partition '%s' in "
-					"section '%s'", dump_fs->partition,
-					name);
-			}
-			return rc;
-		}
+	/* Add data needed to convert fs dump job to IPL job */
+	rc = misc_check_readable_file(FSDUMP_PART_IMAGE);
+	if (rc) {
+		error_text("Need external file '%s' for partition dump",
+			   FSDUMP_PART_IMAGE);
+		return rc;
 	}
-	return 0;
+	dump->image = misc_strdup(FSDUMP_PART_IMAGE);
+	if (dump->image == NULL)
+		return -1;
+	dump->image_addr = DEFAULT_IMAGE_ADDRESS;
+
+	/* Ramdisk is no longer required with new initramfs dump system */
+	if (misc_check_readable_file(FSDUMP_PART_RAMDISK))
+		dump->ramdisk = NULL;
+	else {
+		dump->ramdisk = misc_strdup(FSDUMP_PART_RAMDISK);
+		if (dump->ramdisk == NULL)
+			return -1;
+		dump->ramdisk_addr = UNSPECIFIED_ADDRESS;
+	}
+
+	dump->parm_addr = UNSPECIFIED_ADDRESS;
+	return finalize_dump_address_data(dump, name);
 }
 
 
@@ -544,13 +904,6 @@ check_job_menu_data(struct job_menu_data* menu)
 			if (rc)
 				return rc;
 			break;
-		case job_dump_fs:
-			rc = check_job_dump_fs_data(
-					&menu->entry[i].data.dump_fs,
-					menu->entry[i].name);
-			if (rc)
-				return rc;
-			break;
 		case job_print_usage:
 		case job_print_version:
 		case job_segment:
@@ -606,7 +959,7 @@ check_job_ipl_tape_data(struct job_ipl_tape_data *ipl, char* name)
 			return rc;
 		}
 	}
-	return 0;
+	return finalize_ipl_tape_address_data(ipl, name);
 }
 
 static int
@@ -671,10 +1024,11 @@ check_job_mvdump_data(struct job_mvdump_data* dump, char* name)
 	return 0;
 }
 
+
 static int
 check_job_data(struct job_data* job)
 {
-	int rc;
+	int rc = -1;
 
 	/* Check for missing information */
 	if (job->target.bootmap_dir != NULL) {
@@ -694,26 +1048,29 @@ check_job_data(struct job_data* job)
 	switch (job->id) {
 	case job_print_usage:
 	case job_print_version:
+		rc = 0;
 		break;
 	case job_ipl:
-		return check_job_ipl_data(&job->data.ipl, job->name);
+		rc = check_job_ipl_data(&job->data.ipl, job->name);
+		break;
 	case job_menu:
-		return check_job_menu_data(&job->data.menu);
+		rc = check_job_menu_data(&job->data.menu);
+		break;
 	case job_segment:
-		return check_job_segment_data(&job->data.segment, job->name);
+		rc = check_job_segment_data(&job->data.segment, job->name);
+		break;
 	case job_dump_partition:
-		return check_job_dump_data(&job->data.dump, job->name);
-	case job_dump_fs:
-		return check_job_dump_fs_data(&job->data.dump_fs, job->name);
+		rc = check_job_dump_data(&job->data.dump, job->name);
+		break;
 	case job_ipl_tape:
-		return check_job_ipl_tape_data(&job->data.ipl_tape, job->name);
+		rc = check_job_ipl_tape_data(&job->data.ipl_tape, job->name);
+		break;
 	case job_mvdump:
-		return check_job_mvdump_data(&job->data.mvdump, job->name);
+		rc = check_job_mvdump_data(&job->data.mvdump, job->name);
 	}
-	return 0;
+	return rc;
 }
 
-
 static int
 extract_address (char* string, address_t* address)
 {
@@ -809,7 +1166,7 @@ get_parmline(char* filename, char* line, char** parmline, address_t* address,
 	int to;
 	int got_lf;
 
-	addr = DEFAULT_PARMFILE_ADDRESS;
+	addr = UNSPECIFIED_ADDRESS;
 	if (filename != NULL) {
 		/* Need a filename copy to be able to change it */
 		filename = misc_strdup(filename);
@@ -856,6 +1213,17 @@ get_parmline(char* filename, char* line, char** parmline, address_t* address,
 			return -1;
 
 	} else result = NULL;
+	/* Check for maximum length */
+	if (result) {
+		len = strlen(result);
+		if (len > MAXIMUM_PARMLINE_SIZE) {
+			error_text("The length of the parameters line "
+				   "(%d bytes) exceeds the allowed maximum "
+				   "(%d bytes)", len, MAXIMUM_PARMLINE_SIZE);
+			free(result);
+			return -1;
+		}
+	}
 	*parmline = result;
 	*address = addr;
 	return 0;
@@ -875,10 +1243,10 @@ type_from_target(char *target, disk_type_t *type)
 		*type = disk_type_fba;
 		return 0;
 	case target_type_ldl:
-		*type = disk_type_eckd_classic;
+		*type = disk_type_eckd_ldl;
 		return 0;
 	case target_type_cdl:
-		*type = disk_type_eckd_compatible;
+		*type = disk_type_eckd_cdl;
 		return 0;
 	default:
 		return -1;
@@ -953,7 +1321,18 @@ get_job_from_section_data(char* data[], struct job_data* job, char* section)
 			if (extract_address(job->data.ipl.ramdisk,
 					    &job->data.ipl.ramdisk_addr)) {
 				job->data.ipl.ramdisk_addr =
-					DEFAULT_RAMDISK_ADDRESS;
+					UNSPECIFIED_ADDRESS;
+			}
+		}
+		/* Fill in kdump */
+		if (data[(int) scan_keyword_kdump] != NULL) {
+			if (strcmp(data[(int) scan_keyword_kdump],
+				   "auto") == 0) {
+				job->data.ipl.is_kdump = 1;
+			} else {
+				error_reason("Invalid kdump setting '%s'",
+					     data[(int) scan_keyword_kdump]);
+				return -1;
 			}
 		}
 		break;
@@ -982,14 +1361,16 @@ get_job_from_section_data(char* data[], struct job_data* job, char* section)
 		if (rc)
 			return rc;
 		/* Fill in name and address of ramdisk file */
-		job->data.ipl_tape.ramdisk_addr = DEFAULT_RAMDISK_ADDRESS;
 		if (data[(int) scan_keyword_ramdisk] != NULL) {
 			job->data.ipl_tape.ramdisk =
 				misc_strdup(data[(int) scan_keyword_ramdisk]);
 			if (job->data.ipl_tape.ramdisk == NULL)
 				return -1;
-			extract_address(job->data.ipl_tape.ramdisk,
-					&job->data.ipl_tape.ramdisk_addr);
+			if (extract_address(job->data.ipl_tape.ramdisk,
+					    &job->data.ipl_tape.ramdisk_addr)) {
+				job->data.ipl_tape.ramdisk_addr =
+					UNSPECIFIED_ADDRESS;
+			}
 		}
 		break;
 	case section_segment:
@@ -1023,46 +1404,14 @@ get_job_from_section_data(char* data[], struct job_data* job, char* section)
 			job->data.dump.mem =
 				(job->data.dump.mem + MEGABYTE_MASK) &
 					~MEGABYTE_MASK;
-			printf("Setting dump size limit to %lldMB\n",
+			printf("Found specification of dump size limit "
+			       "to %lldMB\n",
 			       (unsigned long long) job->data.dump.mem /
 			       		(1024LL * 1024LL));
 		}
 		else
 			job->data.dump.mem = -1LL;
 		break;
-	case section_dumpfs:
-		/* DUMP TO FILESYSTEM job */
-		job->id = job_dump_fs;
-		/* Fill in name of bootmap directory */
-		job->target.bootmap_dir =
-			misc_strdup(data[(int) scan_keyword_target]);
-		if (job->target.bootmap_dir == NULL)
-			return -1;
-		/* Fill in partition name */
-		job->data.dump_fs.partition =
-			misc_strdup(data[(int) scan_keyword_dumptofs]);
-		if (job->data.dump_fs.partition == NULL)
-			return -1;
-		/* Check for mem size specification */
-		if (extract_memsize(job->data.dump_fs.partition,
-				    &job->data.dump_fs.mem) == 0) {
-			/* Ensure megabyte alignment of size */
-			job->data.dump_fs.mem =
-				(job->data.dump_fs.mem + MEGABYTE_MASK) &
-					~MEGABYTE_MASK;
-			printf("Setting dump size limit to %lldMB\n",
-			       (unsigned long long) job->data.dump_fs.mem /
-			       		(1024LL * 1024LL));
-		} else
-			job->data.dump_fs.mem = -1LL;
-		/* Fill in parmline */
-		rc = get_parmline(data[(int) scan_keyword_parmfile],
-				  data[(int) scan_keyword_parameters],
-				  &job->data.dump_fs.parmline,
-				  &job->data.dump_fs.parm_addr, section);
-		if (rc)
-			return rc;
-		break;
 	case section_mvdump:
 		/* DUMP TO MULTI-VOLUME job */
 		job->id = job_mvdump;
@@ -1078,7 +1427,8 @@ get_job_from_section_data(char* data[], struct job_data* job, char* section)
 			job->data.mvdump.mem =
 				(job->data.mvdump.mem + MEGABYTE_MASK) &
 					~MEGABYTE_MASK;
-			printf("Setting dump size limit to %lldMB\n",
+			printf("Found specification of dump size limit "
+			       "to %lldMB\n",
 			       (unsigned long long) job->data.mvdump.mem /
 			       		(1024LL * 1024LL));
 		} else
@@ -1118,6 +1468,8 @@ get_menu_job(struct scan_token* scan, char* menu, struct job_data* job)
 	/* Count menu entries, find default entry and target directory */
 	job->data.menu.num = 0;
 	job->data.menu.default_pos = -1;
+	job->data.menu.prompt = MENU_DEFAULT_PROMPT;
+	job->data.menu.timeout = MENU_DEFAULT_TIMEOUT;
 	for (i=index+1; (scan[i].id != scan_id_empty) &&
 			(scan[i].id != scan_id_section_heading) &&
 			(scan[i].id != scan_id_menu_heading); i++) {
@@ -1244,13 +1596,6 @@ get_menu_job(struct scan_token* scan, char* menu, struct job_data* job)
 				memset((void *) &temp_job->data.ipl, 0,
 				       sizeof(struct job_ipl_data));
 				break;
-			case job_dump_fs:
-				job->data.menu.entry[current].id = job_dump_fs;
-				job->data.menu.entry[current].data.dump_fs =
-					temp_job->data.dump_fs;
-				memset((void *) &temp_job->data.dump_fs, 0,
-				       sizeof(struct job_dump_fs_data));
-				break;
 			default:
 				error_reason("Section '%s' cannot be included "
 					     "in menu '%s'", section, menu);
@@ -1266,8 +1611,6 @@ get_menu_job(struct scan_token* scan, char* menu, struct job_data* job)
 }
 
 
-#define DEFAULTBOOT_SECTION	"defaultboot"
-
 static int
 get_default_section(struct scan_token* scan, char** section, int* is_menu)
 {
@@ -1363,19 +1706,6 @@ get_section_job(struct scan_token* scan, char* section, struct job_data* job,
 			if (buffer == NULL)
 				return -1;
 			break;
-		case job_dump_fs:
-			if (job->data.dump_fs.parmline == NULL)
-				buffer = misc_strdup(extra_parmline);
-			else {
-				buffer = append_parmline(
-					job->data.dump_fs.parmline,
-					extra_parmline);
-				free(job->data.dump_fs.parmline);
-			}
-			job->data.dump_fs.parmline = buffer;
-			if (buffer == NULL)
-				return -1;
-			break;
 		case job_segment:
 			error_reason("Option 'parameters' cannot be used with "
 				     "section '%s'", section);
@@ -1399,6 +1729,7 @@ static int
 get_job_from_config_file(struct command_line* cmdline, struct job_data* job)
 {
 	struct scan_token* scan;
+	struct scan_token* new_scan;
 	char* filename;
 	char* source;
 	int rc;
@@ -1424,6 +1755,23 @@ get_job_from_config_file(struct command_line* cmdline, struct job_data* job)
 		error_text("Config file '%s'", filename);
 		return rc;
 	}
+	if ((cmdline->menu == NULL) && (cmdline->section == NULL)) {
+		rc = scan_check_defaultboot(scan);
+		if (rc < 0) {
+			error_text("Config file '%s'", filename);
+			scan_free(scan);
+			return rc;
+		}
+		if (rc == 1) {
+			new_scan = scan_build_automenu(scan);
+			scan_free(scan);
+			if (new_scan == NULL) {
+				error_text("Config file '%s'", filename);
+				return -1;
+			}
+			scan = new_scan;
+		}
+	}
 	rc = scan_check(scan);
 	if (rc) {
 		error_text("Config file '%s'", filename);
