@@ -1,14 +1,15 @@
 /*
- * s390-tools/zipl/src/scan.c
- *   Scanner for zipl.conf configuration files
+ * zipl - zSeries Initial Program Loader tool
  *
- * Copyright IBM Corp. 2001, 2009.
+ * Scanner for zipl.conf configuration files
+ *
+ * Copyright IBM Corp. 2001, 2017
+ *
+ * s390-tools is free software; you can redistribute it and/or modify
+ * it under the terms of the MIT license. See LICENSE for details.
  *
- * Author(s): Carsten Otte <cotte@de.ibm.com>
- *            Peter Oberparleiter <Peter.Oberparleiter@de.ibm.com>
  */
 
-#include "scan.h"
 
 /* Need ISOC99 function isblank() in ctype.h */
 #ifndef __USE_ISOC99
@@ -16,14 +17,18 @@
 #endif
 
 #include <ctype.h>
-#include <string.h>
-#include <stdio.h>
-#include <stdlib.h>
 #include <errno.h>
 #include <limits.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
 
-#include "misc.h"
+#include "lib/util_base.h"
+
+#include "boot.h"
 #include "error.h"
+#include "misc.h"
+#include "scan.h"
 
 
 /* Determines which keyword may be present in which section */
@@ -32,32 +37,45 @@ enum scan_key_state scan_key_table[SCAN_SECTION_NUM][SCAN_KEYWORD_NUM] = {
  *	 ult  to   tofs e    mete file isk  ent  et   pt   out  ultm      dump
  *			     rs                                 enu
  *
- *       targ targ targ targ targ
- *       etba etty etge etbl etof
- *       se   pe   omet ocks fset
+ *       targ targ targ targ targ defa kdum
+ *       etba etty etge etbl etof ulta p
+ *       se   pe   omet ocks fset uto
  *                 ry   ize
  */
-/* defaultboot	*/
-	{opt, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, opt, inv, inv,
-	 inv, inv, inv, inv, inv},
+/* default auto */
+	{opt, inv, inv, inv, inv, inv, inv, inv, req, opt, opt, inv, inv, inv,
+	 opt, opt, opt, opt, opt, opt, inv},
+/* default menu */
+	{inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, req, inv, inv,
+	 inv, inv, inv, inv, inv, inv, inv},
+/* default section */
+	{req, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv,
+	 inv, inv, inv, inv, inv, inv, inv},
 /* ipl		*/
 	{inv, inv, inv, req, opt, opt, opt, inv, req, inv, inv, inv, inv, inv,
-	 opt, opt, opt, opt, opt},
+	 opt, opt, opt, opt, opt, inv, opt},
 /* segment load */
 	{inv, inv, inv, inv, inv, inv, inv, req, req, inv, inv, inv, inv, inv,
-	 inv, inv, inv, inv, inv},
+	 inv, inv, inv, inv, inv, inv, inv},
 /* part dump	*/
 	{inv, req, inv, inv, inv, inv, inv, inv, opt, inv, inv, inv, inv, inv,
-	 inv, inv, inv, inv, inv},
+	 inv, inv, inv, inv, inv, inv, inv},
 /* fs dump	*/
 	{inv, inv, req, inv, opt, opt, inv, inv, req, inv, inv, inv, inv, inv,
-	 inv, inv, inv, inv, inv},
+	 inv, inv, inv, inv, inv, inv, inv},
 /* ipl tape	*/
 	{inv, inv, inv, req, opt, opt, opt, inv, inv, inv, inv, inv, req, inv,
-	 inv, inv, inv, inv, inv},
+	 inv, inv, inv, inv, inv, inv, inv},
 /* multi volume dump	*/
 	{inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, req,
-	 inv, inv, inv, inv, inv}
+	 inv, inv, inv, inv, inv, inv, inv}
+};
+
+/* Determines which keyword may be present in a menu section */
+enum scan_key_state scan_menu_key_table[SCAN_KEYWORD_NUM] = {
+/* menu section */
+	 opt, inv, inv, inv, inv, inv, inv, inv, req, opt, opt, inv, inv, inv,
+	 opt, opt, opt, opt, opt, inv, inv
 };
 
 /* Mapping of keyword IDs to strings */
@@ -83,7 +101,16 @@ static const struct {
 	{ "target", scan_keyword_target},
 	{ "prompt", scan_keyword_prompt},
 	{ "timeout", scan_keyword_timeout},
-	{ "tape", scan_keyword_tape}
+	{ "tape", scan_keyword_tape},
+	{ "kdump", scan_keyword_kdump},
+};
+
+/* List of keywords that are used without an assignment */
+static const struct {
+	char* keyword;
+	enum scan_keyword_id id;
+} keyword_only_list[] = {
+	{ "defaultauto", scan_keyword_defaultauto}
 };
 
 /* Retrieve name of keyword identified by ID. */
@@ -92,9 +119,14 @@ scan_keyword_name(enum scan_keyword_id id)
 {
 	unsigned int i;
 
-	for (i=0; i < sizeof(keyword_list) / sizeof(keyword_list[0]); i++)
+	for (i=0; i < ARRAY_SIZE(keyword_list); i++) {
 		if (id == keyword_list[i].id)
 			return keyword_list[i].keyword;
+	}
+	for (i=0; i < ARRAY_SIZE(keyword_only_list); i++) {
+		if (id == keyword_only_list[i].id)
+			return keyword_only_list[i].keyword;
+	}
 	return "<unknown>";
 }
 
@@ -366,7 +398,7 @@ scan_keyword(struct misc_file_buffer* file, enum scan_keyword_id* id, int line)
 {
 	unsigned int i;
 
-	for (i=0; i < sizeof(keyword_list) / sizeof(keyword_list[0]); i++)
+	for (i=0; i < ARRAY_SIZE(keyword_list); i++)
 		if (match_keyword(file, keyword_list[i].keyword) == 0) {
 			file->pos += strlen(keyword_list[i].keyword);
 			*id = keyword_list[i].id;
@@ -408,6 +440,42 @@ scan_keyword_assignment(struct misc_file_buffer* file, struct scan_token* token,
 }
 
 
+static int
+scan_keyword_only(struct misc_file_buffer* file, enum scan_keyword_id* id,
+		  int line)
+{
+	unsigned int i;
+
+	for (i=0; i < ARRAY_SIZE(keyword_only_list); i++)
+		if (match_keyword(file, keyword_only_list[i].keyword) == 0) {
+			file->pos += strlen(keyword_only_list[i].keyword);
+			*id = keyword_only_list[i].id;
+			return 0;
+		}
+	error_reason("Line %d: unknown keyword", line);
+	return -1;
+}
+
+
+static int
+scan_keyword_only_statement(struct misc_file_buffer* file,
+			    struct scan_token* token, int line)
+{
+	int rc;
+
+	rc = scan_keyword_only(file, &token->content.keyword.keyword, line);
+	if (rc)
+		return rc;
+	if (skip_trailing_blanks(file)) {
+		error_reason("Line %d: unexpected characters at end of line",
+			     line);
+		return -1;
+	}
+	token->id = scan_id_keyword_only;
+	token->line = line;
+	return 0;
+}
+
 
 static int
 search_line_for(struct misc_file_buffer* file, int search)
@@ -496,6 +564,7 @@ scan_file(const char* filename, struct scan_token** token)
 		misc_free_file_buffer(&file);
 		return -1;
 	}
+	memset(array, 0, size * sizeof(struct scan_token));
 	line = 1;
 	while ((size_t) file.pos < file.length) {
 		skip_blanks(&file);
@@ -530,9 +599,8 @@ scan_file(const char* filename, struct scan_token** token)
 					rc = scan_keyword_assignment(
 						&file, &array[pos++], line);
 			} else {
-				error_reason("Line %d: unrecognized directive",
-					     line);
-				rc = -1;
+				rc = scan_keyword_only_statement(&file,
+							&array[pos++], line);
 			}
 		}
 		if (rc)
@@ -548,7 +616,7 @@ scan_file(const char* filename, struct scan_token** token)
 				rc = -1;
 				break;
 			}
-			memset(buffer, 0, size);
+			memset(buffer, 0, size * sizeof(struct scan_token));
 			memcpy(buffer, array, pos*sizeof(struct scan_token));
 			free(array);
 			array = buffer;
@@ -638,8 +706,6 @@ scan_get_target_type(char *type)
 	return target_type_invalid;
 }
 
-#define MAX(a,b)	((a)>(b)?(a):(b))
-
 /* Check section data for correctness. KEYWORD[keyword_id] defines whether
  * a keyword is present, LINE[keyword_id] specifies in which line a keyword
  * was found or NULL when specified on command line, NAME specifies the
@@ -667,8 +733,9 @@ scan_check_section_data(char* keyword[], int* line, char* name,
 			*type = section_dump;
 			main_keyword = scan_keyword_name(scan_keyword_dumpto);
 		} else if (keyword[(int) scan_keyword_dumptofs]) {
-			*type = section_dumpfs;
-			main_keyword = scan_keyword_name(scan_keyword_dumptofs);
+			error_reason("Option dumptofs is deprecated, "
+				     "use dumpto instead");
+			return -1;
 		} else if (keyword[(int) scan_keyword_mvdump]) {
 			*type = section_mvdump;
 			main_keyword = scan_keyword_name(scan_keyword_mvdump);
@@ -743,26 +810,6 @@ scan_check_section_data(char* keyword[], int* line, char* name,
 			return -1;
 		}
 	}
-	/* Additional check needed for defaultboot */
-	if (*type == section_defaultboot) {
-		if ((keyword[(int) scan_keyword_default] != 0) &&
-		    (keyword[(int) scan_keyword_defaultmenu] != 0)) {
-			error_reason("Line %d: Only one of keywords 'default' "
-				     "and 'defaultmenu' allowed in section "
-				     "'defaultboot'",
-				     MAX(line[(int) scan_keyword_default],
-					 line[(int) scan_keyword_defaultmenu]));
-			return -1;
-		}
-		if ((keyword[(int) scan_keyword_default] == 0) &&
-		    (keyword[(int) scan_keyword_defaultmenu] == 0)) {
-			error_reason("Line %d: Section 'defaultboot' requires "
-				     "one of keywords 'default' and "
-				     "'defaultmenu'", section_line);
-			return -1;
-		}
-		
-	}
 	return 0;
 }
 
@@ -780,6 +827,25 @@ check_blocksize(int size)
 	return -1;
 }
 
+static int
+scan_count_target_keywords(char* keyword[])
+{
+	int num = 0;
+
+	if (keyword[(int) scan_keyword_target])
+		num++;
+	if (keyword[(int) scan_keyword_targetbase])
+		num++;
+	if (keyword[(int) scan_keyword_targettype])
+		num++;
+	if (keyword[(int) scan_keyword_targetgeometry])
+		num++;
+	if (keyword[(int) scan_keyword_targetblocksize])
+		num++;
+	if (keyword[(int) scan_keyword_targetoffset])
+		num++;
+	return num;
+}
 
 int
 scan_check_target_data(char* keyword[], int* line)
@@ -924,6 +990,114 @@ scan_check_target_data(char* keyword[], int* line)
 	return 0;
 }
 
+
+static int
+scan_is_delimiter(enum scan_id id)
+{
+	return id == scan_id_empty || id == scan_id_section_heading ||
+	       id == scan_id_menu_heading;
+}
+
+
+/*
+ * Copy data from scan array to keyword/number assignment array. Return 0
+ * on success, non-zero otherwise.
+ *
+ * If keyword X is specified in the section, keyword[X] contains the value of
+ * that keyword and keyword_line[x] the line in which the keyword was defined.
+ * Otherwise keyword[X] contains NULL.
+ * If number assignment X is specified in the section, num[X - 1] contains
+ * the value of that assignment and num_line the line in which the assignment
+ * was found. Otherwise num[X - 1] contains NULL. If num is NULL, an error
+ * is reported if a number assignment is found.
+ */
+static int
+scan_get_section_keywords(struct scan_token* scan, int* index, char* name,
+			  char* keyword[], int keyword_line[], char** num,
+			  int* num_line)
+{
+	int i;
+	int id;
+	int key;
+	int line;
+	int number;
+	char* value;
+
+	/* Initialize array data. */
+	for (i = 0; i < SCAN_KEYWORD_NUM; i++) {
+		keyword[i] = NULL;
+		keyword_line[i] = 0;
+	}
+	if (num) {
+		for (i = 0; i < BOOT_MENU_ENTRIES; i++) {
+			num[i] = NULL;
+			num_line[i] = 0;
+		}
+	}
+	/* Extract data for keyword and number assignments. */
+	for (i = *index + 1; !scan_is_delimiter(scan[i].id); i++) {
+		id = (int) scan[i].id;
+		line = scan[i].line;
+		/* Handle number assignments first. */
+		if (id == scan_id_number_assignment) {
+			number = scan[i].content.number.number;
+			/* Check if number assignments are expected. */
+			if (!num) {
+				error_reason("Line %d: number assignment not "
+					     "allowed in section '%s'",
+					     line, name);
+				return -1;
+			}
+			/* Check for valid numbers. */
+			if (number <= 0) {
+				error_reason("Line %d: position must be "
+					     "greater than zero",
+					     line);
+                                return -1;
+
+			}
+			if (number > BOOT_MENU_ENTRIES) {
+				error_reason("Line %d: position must not "
+					     "exceed %d", line,
+					     BOOT_MENU_ENTRIES);
+                                return -1;
+			}
+			/* Rule 10 */
+			if (num[number - 1]) {
+				error_reason("Line %d: position %d already "
+					     "specified", line, number);
+				return -1;
+			}
+			num[number - 1] = scan[i].content.number.value;
+			num_line[number - 1] = line;
+			continue;
+		}
+		/* Handle keyword assignments. */
+		if (id == scan_id_keyword_assignment) {
+			key = scan[i].content.keyword.keyword;
+			value = scan[i].content.keyword.value;
+		} else if (id == scan_id_keyword_only) {
+			key = scan[i].content.keyword_only.keyword;
+			/* Define a dummy value. */
+			value = "";
+		} else {
+			continue;
+		}
+		if (keyword[key]) {
+			/* Rule 5 */
+			error_reason("Line %d: keyword '%s' already specified",
+				     line, scan_keyword_name(key));
+			return -1;
+		}
+		keyword[key] = value;
+		keyword_line[key] = line;
+	}
+	*index = i;
+
+	return 0;
+}
+
+
 /* Check section at INDEX for compliance with config file rules. Upon success,
  * return zero and advance INDEX to point to the end of the section. Return
  * non-zero otherwise. */
@@ -933,64 +1107,30 @@ check_section(struct scan_token* scan, int* index)
 	char* name;
 	char* keyword[SCAN_KEYWORD_NUM];
 	int keyword_line[SCAN_KEYWORD_NUM];
-	enum scan_keyword_id key_id;
 	enum scan_section_type type;
 	int line;
-	int i;
 	int rc;
 	
-	i = *index;
-	name = scan[i].content.section.name;
+	name = scan[*index].content.section.name;
 	/* Ensure unique section names */
-	line = scan_find_section(scan, name, scan_id_section_heading, i+1);
+	line = scan_find_section(scan, name, scan_id_section_heading, *index+1);
 	if (line >= 0) {
 		error_reason("Line %d: section name '%s' already specified",
 			     scan[line].line, name);
 		return -1;
 	}
-	/* Check for defaultboot section */
-	if (strcmp(name, "defaultboot") == 0)
-		type = section_defaultboot;
+	line = scan[*index].line;
+	/* Get keyword data and advance index to end of section */
+	rc = scan_get_section_keywords(scan, index, name, keyword, keyword_line,
+				       NULL, NULL);
+	if (rc)
+		return rc;
+	/* Check already done in scan_check_defaultboot */
+	if (strcmp(name, DEFAULTBOOT_SECTION) == 0)
+		return 0;
 	else
 		type = section_invalid;
-	memset(keyword, 0, sizeof(keyword));
-	line = scan[i].line;
-	/* Account for keywords */
-	for (i++; (int) scan[i].id != 0; i++) {
-		if ((scan[i].id == scan_id_section_heading) ||
-		    (scan[i].id == scan_id_menu_heading))
-			break;
-		else if (scan[i].id == scan_id_keyword_assignment) {
-			key_id = scan[i].content.keyword.keyword;
-			keyword_line[key_id] = scan[i].line;
-			if (keyword[(int) key_id] != NULL) {
-				/* Rule 5 */
-				error_reason("Line %d: keyword '%s' already "
-					     "specified", scan[i].line,
-					     scan_keyword_name(key_id));
-				return -1;
-			}
-			keyword[(int) key_id] = scan[i].content.keyword.value;
-			/* Check default keyword value */
-			if (key_id == scan_keyword_default) {
-				/* Rule 8 */
-				if (scan_find_section(scan,
-				     scan[i].content.keyword.value,
-				     scan_id_section_heading, 0) < 0) {
-					error_reason("Line %d: no such section "
-						"'%s'", scan[i].line,
-						scan[i].content.keyword.value);
-					return -1;
-				}
-			}
-		} else if (scan[i].id == scan_id_number_assignment) {
-			/* Rule 2 */
-			error_reason("Line %d: number assignment not allowed "
-				     "in section '%s'", scan[i].line,
-				     name);
-			return -1;
-		}
-	}
+	/* Check section data */
 	rc = scan_check_section_data(keyword, keyword_line, name, line, &type);
 	if (rc)
 		return rc;
@@ -998,29 +1138,42 @@ check_section(struct scan_token* scan, int* index)
 	rc = scan_check_target_data(keyword, keyword_line);
 	if (rc)
 		return rc;
-	/* Advance index to end of section */
-	*index = i - 1;
 	return 0;
 }
 
 
 static int
-find_num_assignment(struct scan_token* scan, int num, int offset)
+check_menu_keyword_data(char* keyword[], int* line, char* name,
+			int section_line)
 {
 	int i;
 
-	for (i=offset; (int) scan[i].id != 0; i++) {
-		if (scan[i].id == scan_id_number_assignment) {
-			if (scan[i].content.number.number == num)
-				return i;
-		} else if ((scan[i].id == scan_id_section_heading) ||
-			   (scan[i].id == scan_id_menu_heading))
+	for (i = 0; i < SCAN_KEYWORD_NUM; i++) {
+		switch (scan_menu_key_table[i]) {
+		case inv:
+			if (!keyword[i])
+				break;
+			error_reason("Line %d: keyword '%s' not allowed in "
+				     "menu section '%s'", line[i],
+				     scan_keyword_name(
+					(enum scan_keyword_id) i), name);
+				return -1;
 			break;
+		case req:
+			if (keyword[i])
+				break;
+			error_reason("Line %d: missing keyword '%s' "
+				     "in menu section '%s'", section_line,
+				     scan_keyword_name(
+					(enum scan_keyword_id) i), name);
+			return -1;
+		default:
+			break;
+		}
 	}
-	return -1;
+	return 0;
 }
 
-
 /* Check menu section at INDEX for compliance with config file rules. Upon
  * success, return zero and advance INDEX to point to the end of the section.
  * Return non-zero otherwise. */
@@ -1029,238 +1182,139 @@ check_menu(struct scan_token* scan, int* index)
 {
 	char* keyword[SCAN_KEYWORD_NUM];
 	int keyword_line[SCAN_KEYWORD_NUM];
-	enum scan_keyword_id key_id;
-	char* name;
-	char* str;
-	long pos;
-	int line;
+	char* num[BOOT_MENU_ENTRIES];
+	int num_line[BOOT_MENU_ENTRIES];
+	char* menu_name;
+	int menu_line;
+	char dummy;
 	int i;
-	int j;
-	int is_num;
-	int is_target;
-	int is_default;
-	int is_prompt;
-	int is_timeout;
 	int rc;
+	int number;
+	int num_configs;
 
-	i = *index;
-	name = scan[i].content.menu.name;
+	menu_name = scan[*index].content.menu.name;
+	menu_line = scan[*index].line;
 	/* Rule 15 */
-	line = scan_find_section(scan, name, scan_id_menu_heading, i+1);
-	if (line >= 0) {
+	i = scan_find_section(scan, menu_name, scan_id_menu_heading,
+			      *index + 1);
+	if (i >= 0) {
 		error_reason("Line %d: menu name '%s' already specified",
-			     scan[line].line, name);
+			     scan[i].line, menu_name);
 		return -1;
 	}
-	line = scan[i].line;
-	is_num = 0;
-	is_target = 0;
-	is_default = 0;
-	is_prompt = 0;
-	is_timeout = 0;
-	/* Account for keywords */
-	for (i++; (int) scan[i].id != 0; i++) {
-		if ((scan[i].id == scan_id_section_heading) ||
-		    (scan[i].id == scan_id_menu_heading))
-			break;
-		else if (scan[i].id == scan_id_keyword_assignment) {
-			key_id = scan[i].content.keyword.keyword;
-			switch (key_id) {
-			case scan_keyword_target:
-				if (is_target) {
-					error_reason("Line %d: keyword '%s' "
-						     "already specified",
-						     scan[i].line,
-						     scan_keyword_name(key_id));
-					return -1;
-				}
-				is_target = 1;
-				break;
-			case scan_keyword_targetbase:
-			case scan_keyword_targettype:
-			case scan_keyword_targetgeometry:
-			case scan_keyword_targetblocksize:
-			case scan_keyword_targetoffset:
-				key_id = scan[i].content.keyword.keyword;
-				keyword_line[key_id] = scan[i].line;
-				/* Rule 5 */
-				if (keyword[(int) key_id] != NULL) {
-					error_reason("Line %d: keyword '%s' "
-						"already specified",
-						scan[i].line,
-						scan_keyword_name(key_id));
-					return -1;
-				}
-				keyword[(int) key_id] =
-					scan[i].content.keyword.value;
-				break;
-			case scan_keyword_default:
-				if (is_default) {
-					error_reason("Line %d: keyword '%s' "
-						     "already specified",
-						     scan[i].line,
-						     scan_keyword_name(key_id));
-					return -1;
-				}
-				is_default = 1;
-				/* Rule 16 */
-				str = scan[i].content.keyword.value;
-				/* Skip leading blanks */
-				for (; isspace(*str); str++) ;
-				if (*str == 0) {
-					error_reason("Line %d: default "
-						     "position must be a "
-						     "number", scan[i].line);
-					return -1;
-				}
-				pos = strtol(str, &str, 10);
-				/* Skip blanks after number */
-				for (; isspace(*str); str++) ;
-				if (*str != 0) {
-					error_reason("Line %d: default "
-						     "position must be a "
-						     "number", scan[i].line);
-					return -1;
-				}
-				if (pos <= 0) {
-					error_reason("Line %d: default "
-						     "position must be greater "
-						     "than zero",
-						     scan[i].line);
-					return -1;
-				}
-				if ((pos == LONG_MAX) && (errno == ERANGE)) {
-					error_reason("Line %d: default "
-						     "position too large",
-						     scan[i].line);
-					return -1;
-				}
-				j = find_num_assignment(scan, pos, *index+1);
-				if (j < 0) {
-					error_reason("Line %d: menu "
-						     "position %ld not "
-						     "defined",
-						     scan[i].line, pos);
-					return -1;
-				}
-				break;
-			case scan_keyword_prompt:
-				if (is_prompt) {
-					error_reason("Line %d: keyword '%s' "
-						     "already specified",
-						     scan[i].line,
-						     scan_keyword_name(key_id));
-					return -1;
-				}
-				is_prompt = 1;
-				str = scan[i].content.keyword.value;
-				/* Skip leading blanks */
-				for (; isspace(*str); str++) ;
-				if (*str == 0) {
-					error_reason("Line %d: prompt value "
-						     "must be a number",
-						     scan[i].line);
-					return -1;
-				}
-				pos = strtol(str, &str, 10);
-				/* Skip blanks after number */
-				for (; isspace(*str); str++) ;
-				if (*str != 0) {
-					error_reason("Line %d: prompt value "
-						     "must be a number",
-						     scan[i].line);
-					return -1;
-				}
-				break;
-			case scan_keyword_timeout:
-				if (is_timeout) {
-					error_reason("Line %d: keyword '%s' "
-						     "already specified",
-						     scan[i].line,
-						     scan_keyword_name(key_id));
-					return -1;
-				}
-				is_timeout = 1;
-				str = scan[i].content.keyword.value;
-				/* Skip leading blanks */
-				for (; isspace(*str); str++) ;
-				if (*str == 0) {
-					error_reason("Line %d: timeout value "
-						     "must be a number",
-						     scan[i].line);
-					return -1;
-				}
-				pos = strtol(str, &str, 10);
-				/* Skip blanks after number */
-				for (; isspace(*str); str++) ;
-				if (*str != 0) {
-					error_reason("Line %d: timeout value "
-						     "must be a number",
-						     scan[i].line);
-					return -1;
-				}
-				break;
-			default:
-				/* Rule 13 */
-				error_reason("Line %d: keyword '%s' not "
-					     "allowed in menu", scan[i].line,
-					     scan_keyword_name(key_id));
-				return -1;
-			}
-		} else if (scan[i].id == scan_id_number_assignment) {
-			is_num = 1;
-			if (scan[i].content.number.number < 1) {
-				error_reason("Line %d: position must be "
-					     "greater than zero",
-					     scan[i].line);
-				return -1;
-			}
-			/* Rule 10 */
-			j = find_num_assignment(scan,
-						scan[i].content.number.number,
-						i+1);
-			if (j >= 0) {
-				error_reason("Line %d: position %d already "
-					     "specified",
-					     scan[j].line,
-					     scan[i].content.number.number);
-				return -1;
-			}
-			/* Rule 11 */
-			j = scan_find_section(scan,
-					      scan[i].content.number.value,
-					      scan_id_section_heading, 0);
-			if (j < 0) {
-				error_reason("Line %d: section '%s' not found",
-					     scan[i].line,
-					     scan[i].content.number.value);
-				return -1;
-			}
-
+	/* Get keyword and number assignment data */
+	rc = scan_get_section_keywords(scan, index, menu_name, keyword,
+				       keyword_line, num, num_line);
+	if (rc)
+		return rc;
+	/* Check for required and invalid keywords */
+	rc = check_menu_keyword_data(keyword, keyword_line, menu_name,
+				     menu_line);
+	if (rc)
+		return rc;
+	/* Check default value */
+	i = (int) scan_keyword_default;
+	if (keyword[i]) {
+		if (sscanf(keyword[i], "%d %c", &number, &dummy) != 1) {
+			error_reason("Line %d: default position must be a "
+				     "number", keyword_line[i]);
+			return -1;
+		}
+		if (number < 1) {
+			error_reason("Line %d: default position must be "
+				     "greater than zero", keyword_line[i]);
+			return -1;
+		}
+		if (number > BOOT_MENU_ENTRIES) {
+			error_reason("Line %d: default position too large",
+				     keyword_line[i]);
+			return -1;
+		}
+		if (!num[number - 1]) {
+			error_reason("Line %d: menu position %d not defined",
+				     keyword_line[i], number);
+			return -1;
 		}
 	}
-	/* Rule 9 */
-	if (!is_num) {
-		error_reason("Line %d: no boot configuration specified in "
-			     "menu '%s'", line, name);
-		return -1;
+	/* Check prompt value */
+	i = (int) scan_keyword_prompt;
+	if (keyword[i]) {
+		if (sscanf(keyword[i], "%d %c", &number, &dummy) != 1) {
+			error_reason("Line %d: prompt value must be a number",
+				     keyword_line[i]);
+			return -1;
+		}
+		if (number != 0 && number != 1) {
+			error_reason("Line %d: prompt value is out of range "
+				     "(must be 0 or 1)", keyword_line[i]);
+			return -1;
+		}
 	}
-	/* Rule 12 */
-	if (!is_target) {
-		error_reason("Line %d: missing keyword '%s' in menu '%s'",
-			     line, scan_keyword_name(scan_keyword_target),
-			     name);
+	/* Check timeout value */
+	i = (int) scan_keyword_timeout;
+	if (keyword[i]) {
+		if (sscanf(keyword[i], "%d %c", &number, &dummy) != 1) {
+			error_reason("Line %d: timeout value must be a number",
+				     keyword_line[i]);
+			return -1;
+		}
+		if (number < 0) {
+			error_reason("Line %d: timeout value is out of range",
+				     keyword_line[i]);
+			return -1;
+		}
+	}
+	/* Check number assignments */
+	num_configs = 0;
+	for (i = 0; i < BOOT_MENU_ENTRIES; i++) {
+		if (!num[i])
+			continue;
+		if (scan_find_section(scan, num[i], scan_id_section_heading,
+				      0) < 0) {
+			error_reason("Line %d: section '%s' not found",
+				     num_line[i], num[i]);
+			return -1;
+		}
+		num_configs++;
+	}
+	/* Check for boot configurations */
+	if (num_configs == 0) {
+		error_reason("Line %d: no boot configuration specified in "
+			     "menu '%s'", menu_line, menu_name);
 		return -1;
 	}
 	/* Check target data */
 	rc = scan_check_target_data(keyword, keyword_line);
 	if (rc)
 		return rc;
-	/* Advance index to end of menu section */
-	*index = i - 1;
+
 	return 0;
 }
 
+static const struct {
+	char* name;
+	enum scan_id id;
+} id_list[] = {
+	{ "none", scan_id_empty },
+	{ "section heading", scan_id_section_heading },
+	{ "menu heading", scan_id_menu_heading },
+	{ "keyword assignment",  scan_id_keyword_assignment },
+	{ "number assignment", scan_id_number_assignment},
+	{ "keyword", scan_id_keyword_only},
+};
+
+static const char *
+scan_id_name(enum scan_id id)
+{
+	unsigned int i;
+
+	for (i = 0; i < ARRAY_SIZE(id_list); i++) {
+		if (id == id_list[i].id)
+			return id_list[i].name;
+	}
+	return "<unknown>";
+}
+
 
 /* Check scanned tokens for compliance with config file rules. Return zero
  * if config file complies, non-zero otherwise. Rules are:
@@ -1290,46 +1344,476 @@ check_menu(struct scan_token* scan, int* index)
 int
 scan_check(struct scan_token* scan)
 {
-	int in_section;
 	int i;
 	int rc;
 
-	in_section = 0;
-	for (i=0; (int) scan[i].id != 0; i++) {
+	i = 0;
+	while (scan[i].id != scan_id_empty) {
 		switch (scan[i].id) {
 		case scan_id_section_heading:
-			in_section = 1;
 			rc = check_section(scan, &i);
 			if (rc)
 				return rc;
 			break;
 		case scan_id_menu_heading:
-			in_section = 1;
 			rc = check_menu(scan, &i);
 			if (rc)
 				return rc;
 			break;
-		case scan_id_keyword_assignment:
-			/* Rule 1 */
-			if (!in_section) {
-				error_reason("Line %d: keyword assignment not "
-					     "allowed outside of section",
-					     scan[i].line);
-				return -1;
-			}
-			break;
-		case scan_id_number_assignment:
-			/* Rule 1 */
-			if (!in_section) {
-				error_reason("Line %d: number assignment not "
-					     "allowed outside of section",
-					     scan[i].line);
-				return -1;
-			}
-			break;
 		default:
-			break;
+			/* Rule 1 */
+			error_reason("Line %d: %s not allowed outside of "
+				     "section", scan[i].line,
+				     scan_id_name(scan[i].id));
+			return -1;
 		}
 	}
 	return 0;
 }
+
+
+static int
+scan_get_defaultboot_type(char* keyword[], int line[], int section_line,
+			  enum scan_section_type* type)
+{
+	int key1;
+	int key2;
+
+	if (keyword[(int) scan_keyword_defaultauto]) {
+		if (keyword[(int) scan_keyword_defaultmenu]) {
+			key1 = (int) scan_keyword_defaultauto;
+			key2 = (int) scan_keyword_defaultmenu;
+			goto err_too_much;
+		}
+		if (!keyword[(int) scan_keyword_target]) {
+			goto err_no_target;
+		}
+		*type = section_default_auto;
+	} else if (keyword[(int) scan_keyword_target]) {
+		if (keyword[(int) scan_keyword_defaultmenu]) {
+			key1 = (int) scan_keyword_target;
+			key2 = (int) scan_keyword_defaultmenu;
+			goto err_too_much_target;
+		}
+		*type = section_default_auto;
+	} else if (keyword[(int) scan_keyword_defaultmenu]) {
+		if (keyword[(int) scan_keyword_default]) {
+			key1 = (int) scan_keyword_defaultmenu;
+			key2 = (int) scan_keyword_default;
+			goto err_too_much;
+		}
+		*type = section_default_menu;
+	} else if (keyword[(int) scan_keyword_default]) {
+		*type = section_default_section;
+	} else {
+		error_reason("Line %d: Section '%s' requires one of keywords "
+			     "'default', 'defaultmenu', 'defaultauto' or "
+			     "'target'", section_line, DEFAULTBOOT_SECTION);
+		return -1;
+	}
+	return 0;
+
+err_no_target:
+	error_reason("Line %d: Keyword 'target' required in section '%s' when "
+		     "specifying 'defaultauto'", section_line,
+		     DEFAULTBOOT_SECTION);
+	return -1;
+
+err_too_much:
+	error_reason("Line %d: Only one of keywords 'default', 'defaultmenu' "
+		     "and 'defaultauto' allowed in section '%s'",
+		     MAX(line[key1], line[key2]), DEFAULTBOOT_SECTION);
+	return -1;
+
+err_too_much_target:
+	error_reason("Line %d: Only one of keywords 'default', 'defaultmenu' "
+		     "and 'target' allowed in section '%s'",
+		     MAX(line[key1], line[key2]), DEFAULTBOOT_SECTION);
+	return -1;
+}
+
+
+/* scan_check_defaultboot checks the defaultboot section of the configuration
+ * file. It returns
+ *    0   on successful check, NO automenu is to be built
+ *    1   on successful check, the automenu must be built
+ *   -1   on error
+ */
+int
+scan_check_defaultboot(struct scan_token* scan)
+{
+	int i;
+	int j;
+	char* keyword[SCAN_KEYWORD_NUM];
+	int keyword_line[SCAN_KEYWORD_NUM];
+	enum scan_section_type type;
+	int defaultboot_line;
+	int rc;
+
+	/* Check if defaultboot is defined */
+	i = scan_find_section(scan, DEFAULTBOOT_SECTION,
+			      scan_id_section_heading, 0);
+	if (i < 0) {
+		error_reason("No '%s' section found and no section specified "
+			     "on command line", DEFAULTBOOT_SECTION);
+		return -1;
+	}
+	/* Ensure unique section names */
+	j = scan_find_section(scan, DEFAULTBOOT_SECTION,
+			      scan_id_section_heading, i + 1);
+	if (j >= 0) {
+		error_reason("Line %d: section name '%s' already specified",
+			     scan[j].line, DEFAULTBOOT_SECTION);
+		return -1;
+	}
+	defaultboot_line = scan[i].line;
+	/* Get keyword data */
+	rc = scan_get_section_keywords(scan, &i, DEFAULTBOOT_SECTION, keyword,
+				       keyword_line, NULL, NULL);
+	if (rc)
+		return rc;
+	/* Check default keyword value */
+	i = (int) scan_keyword_default;
+	if (keyword[i] && scan_find_section(scan, keyword[i],
+					    scan_id_section_heading, 0) < 0) {
+		error_reason("Line %d: no such section '%s'", keyword_line[i],
+			     keyword[i]);
+		return -1;
+	}
+	/* Determine default boot type */
+	rc = scan_get_defaultboot_type(keyword, keyword_line, defaultboot_line,
+				       &type);
+	if (rc)
+		return rc;
+	/* Check target keywords */
+	if (type == section_default_auto &&
+	    scan_count_target_keywords(keyword) > 0) {
+		rc = scan_check_target_data(keyword, keyword_line);
+		if (rc)
+			return rc;
+	}
+	/* Check remaining section data */
+	rc = scan_check_section_data(keyword, keyword_line, DEFAULTBOOT_SECTION,
+				     defaultboot_line, &type);
+	if (rc)
+		return rc;
+	if (type == section_default_auto)
+		return 1;
+	else
+		return 0;
+}
+
+
+static void
+scan_skip_section(struct scan_token* scan, int* index)
+{
+	do {
+		(*index)++;
+	} while (!scan_is_delimiter(scan[*index].id));
+}
+
+
+static int
+scan_copy_section(struct scan_token* from, struct scan_token* to,
+		  int* index_from, int* index_to)
+{
+	struct scan_token* token;
+
+	do {
+		token = &to[*index_to];
+		*token = from[*index_from];
+		switch (token->id) {
+		case scan_id_section_heading:
+			token->content.section.name =
+				misc_strdup(token->content.section.name);
+			if (!token->content.section.name)
+				return -1;
+			break;
+		case scan_id_menu_heading:
+			token->content.menu.name =
+				misc_strdup(token->content.menu.name);
+			if (!token->content.menu.name)
+				return -1;
+			break;
+		case scan_id_keyword_assignment:
+			token->content.keyword.value =
+				misc_strdup(token->content.keyword.value);
+			if (!token->content.keyword.value)
+				return -1;
+			break;
+		case scan_id_number_assignment:
+			token->content.number.value =
+				misc_strdup(token->content.number.value);
+			if (!token->content.number.value)
+				return -1;
+			break;
+		default:
+			break;
+		}
+		(*index_from)++;
+		(*index_to)++;
+	} while (!scan_is_delimiter(from[*index_from].id));
+
+	return 0;
+}
+
+
+static int
+scan_append_section_heading(struct scan_token* scan, int* index, char* name)
+{
+	scan[*index].id = scan_id_section_heading;
+	scan[*index].line = 0;
+	scan[*index].content.section.name = misc_strdup(name);
+	if (!scan[*index].content.section.name)
+		return -1;
+	(*index)++;
+	return 0;
+}
+
+
+static int
+scan_append_menu_heading(struct scan_token* scan, int* index, char* name)
+{
+	scan[*index].id = scan_id_menu_heading;
+	scan[*index].line = 0;
+	scan[*index].content.menu.name = misc_strdup(name);
+	if (!scan[*index].content.menu.name)
+		return -1;
+	(*index)++;
+	return 0;
+}
+
+
+static int
+scan_append_number_assignment(struct scan_token* scan, int* index, int num,
+			      char* value)
+{
+	scan[*index].id = scan_id_number_assignment;
+	scan[*index].line = 0;
+	scan[*index].content.number.number = num;
+	scan[*index].content.number.value = misc_strdup(value);
+	if (!scan[*index].content.number.value)
+		return -1;
+	(*index)++;
+	return 0;
+}
+
+
+static int
+scan_append_keyword_assignment(struct scan_token* scan, int* index,
+			       enum scan_keyword_id id, char* value)
+{
+	scan[*index].id = scan_id_keyword_assignment;
+	scan[*index].line = 0;
+	scan[*index].content.keyword.keyword = id;
+	scan[*index].content.keyword.value = misc_strdup(value);
+	if (!scan[*index].content.keyword.value)
+		return -1;
+	(*index)++;
+	return 0;
+}
+
+
+static int
+scan_append_target_keywords(struct scan_token* scan, int* index,
+			    char* keyword[])
+{
+	enum scan_keyword_id id[] = {
+		scan_keyword_target,
+		scan_keyword_targetbase,
+		scan_keyword_targettype,
+		scan_keyword_targetgeometry,
+		scan_keyword_targetblocksize,
+		scan_keyword_targetoffset,
+	};
+	unsigned int i;
+	int rc;
+
+	for (i = 0; i < ARRAY_SIZE(id); i++) {
+		if (!keyword[(int) id[i]])
+			continue;
+		rc = scan_append_keyword_assignment(scan, index, id[i],
+						    keyword[(int) id[i]]);
+		if (rc)
+			return rc;
+	}
+
+	return 0;
+}
+
+
+struct scan_token *
+scan_build_automenu(struct scan_token* scan)
+{
+	char* entry[BOOT_MENU_ENTRIES];
+	int num_entries;
+	int default_entry;
+	char* db_keyword[SCAN_KEYWORD_NUM];
+	int db_line[SCAN_KEYWORD_NUM];
+	char* sec_keyword[SCAN_KEYWORD_NUM];
+	int sec_line[SCAN_KEYWORD_NUM];
+	int num_targets;
+	int num_sections;
+	size_t size;
+	struct scan_token* new_scan;
+	int i;
+	int i_new;
+	char* name;
+	char* default_name;
+	int pos;
+
+	/* Find defaultboot */
+	i = scan_find_section(scan, DEFAULTBOOT_SECTION,
+			      scan_id_section_heading, 0);
+	/* Get defaultboot data */
+	if (scan_get_section_keywords(scan, &i, DEFAULTBOOT_SECTION, db_keyword,
+				      db_line, NULL, NULL))
+		return NULL;
+	default_name = db_keyword[(int) scan_keyword_default];
+	num_targets = scan_count_target_keywords(db_keyword);
+	/* Get size of scan array and number of sections */
+	num_sections = 0;
+	for (i = 0; scan[i].id != scan_id_empty; i++) {
+		if (scan[i].id == scan_id_section_heading)
+			num_sections++;
+	}
+	size = /* old scan array + delimiter */     i + 1 +
+	       /* defaultboot heading  */           1 +
+	       /* defaultmenu */                    1 +
+	       /* menu heading  */                  1 +
+	       /* keyword default,prompt,timeout */ 3 +
+	       /* target keywords*/                 num_targets +
+	       /* missing target definitions */     num_sections * num_targets +
+	       /* number assigment  */              num_sections;
+	size *= sizeof(struct scan_token);
+	new_scan = misc_malloc(size);
+	if (!new_scan)
+		return NULL;
+	memset(new_scan, 0, size);
+	/* Fill new array */
+	i = 0;
+	i_new = 0;
+	num_entries = 0;
+	default_entry = -1;
+	memset(entry, 0, sizeof(entry));
+	while (scan[i].id != scan_id_empty) {
+		switch (scan[i].id) {
+		case scan_id_menu_heading:
+			name = scan[i].content.menu.name;
+			/* Abort if automenu name is already in use */
+			if (strcmp(name, SCAN_AUTOMENU_NAME) == 0) {
+				error_reason("Cannot build automenu: menu name "
+					     "'%s' already used",
+					     SCAN_AUTOMENU_NAME);
+				goto err;
+			}
+			/* Menu sections are copied without changes */
+			if (scan_copy_section(scan, new_scan, &i, &i_new))
+				goto err;
+			break;
+		case scan_id_section_heading:
+			name = scan[i].content.section.name;
+			/* Do not copy old defaultboot section */
+			if (strcmp(name, DEFAULTBOOT_SECTION) == 0) {
+				scan_skip_section(scan, &i);
+				break;
+			}
+			/* Get section data but do not advance index */
+			pos = i;
+			if (scan_get_section_keywords(scan, &pos, name,
+						      sec_keyword, sec_line,
+						      NULL, NULL))
+				goto err;
+			/* Copy section contents and advance index */
+			if (scan_copy_section(scan, new_scan, &i, &i_new))
+				goto err;
+			/* Stop here for non-IPL sections. */
+			if (scan_get_section_type(sec_keyword) != section_ipl)
+				break;
+			/* Is there enough room for another section? */
+			if (num_entries == BOOT_MENU_ENTRIES) {
+				error_reason("Cannot build automenu: too many "
+					     "IPL sections defined (max %d)",
+					     BOOT_MENU_ENTRIES);
+				goto err;
+			}
+			/* Determine if this is the default entry */
+			if (default_name && strcmp(default_name, name) == 0)
+				default_entry = num_entries;
+			entry[num_entries++] = name;
+			/* Add missing target parameters if necessary */
+			if (scan_count_target_keywords(sec_keyword) > 0)
+				break;
+			if (scan_append_target_keywords(new_scan, &i_new,
+							db_keyword))
+				goto err;
+			break;
+		default:
+			/* Rule 1 */
+			error_reason("Line %d: %s not allowed outside of "
+				     "section", scan[i].line,
+				     scan_id_name(scan[i].id));
+			goto err;
+		}
+	}
+	if (num_entries == 0) {
+		error_reason("Cannot build automenu: no IPL entries available");
+		goto err;
+	}
+
+	/* Append new defaultboot and automenu sections */
+	/* [defaultboot] */
+	if (scan_append_section_heading(new_scan, &i_new, DEFAULTBOOT_SECTION))
+		goto err;
+	/* defaultmenu=zipl-automatic-menu */
+	if (scan_append_keyword_assignment(new_scan, &i_new,
+				scan_keyword_defaultmenu, SCAN_AUTOMENU_NAME))
+		goto err;
+	/* :zipl-automatic-menu */
+	if (scan_append_menu_heading(new_scan, &i_new, SCAN_AUTOMENU_NAME))
+		goto err;
+	/* default= */
+	if (default_entry >= 0) {
+		char str[20];
+
+		snprintf(str, sizeof(str), "%d", default_entry + 1);
+		if (scan_append_keyword_assignment(new_scan, &i_new,
+						   scan_keyword_default, str))
+			goto err;
+	}
+	/* prompt= */
+	i = (int) scan_keyword_prompt;
+	if (db_keyword[i]) {
+		if (scan_append_keyword_assignment(new_scan, &i_new,
+						   scan_keyword_prompt,
+						   db_keyword[i]))
+			goto err;
+	}
+	/* timeout= */
+	i = (int) scan_keyword_timeout;
+	if (db_keyword[i]) {
+		if (scan_append_keyword_assignment(new_scan, &i_new,
+						   scan_keyword_timeout,
+						   db_keyword[i]))
+			goto err;
+	}
+	/* target= */
+	/* targetbase= */
+	/* targetgeometry= */
+	/* targetblocksize= */
+	/* targetoffset= */
+	if (scan_append_target_keywords(new_scan, &i_new, db_keyword))
+		goto err;
+	/* <num>=<section name>*/
+	for (i = 0; i < num_entries; i++) {
+		if (scan_append_number_assignment(new_scan, &i_new, i + 1,
+						  entry[i]))
+			goto err;
+	}
+
+	return new_scan;
+
+err:
+	scan_free(new_scan);
+	return NULL;
+}
